/*
 * QEMU GTK2 display driver
 * based on SDL driver by Fabrice
 * 
 * Copyright (c) 2005 Jim Brown
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include "vl.h"

#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "fullscreen.h"

/* define our own bitshift enums to allow qemugtk to know difference between left and right alt - something gtk doesnt provide in its modifiers mask. this uses qemu's own modifier_state[] map in order to guess correctly */
typedef enum
{
	gtkshiftleft = 1 << 0,
	gtkshiftright = 1 << 1,
	gtkcontrolleft = 1 << 2,
	gtkcontrolright = 1 << 3,
	gtkaltleft = 1 << 4,
	gtkaltright = 1 << 5,
	gtkcapslock = 1 << 6
} gtk2keymod;


static GtkWidget *window, *swindow;
static GtkWidget *event_port, *menubar;
static GdkImage *image=NULL;
static GdkCursor *invisible_cursor;
static int ox = 0, oy = 0;
static int gui_grab; /* if true, all keyboard/mouse events are grabbed */
static int gui_rgrab; /* if true, all keyboard/mouse events are grabbed */
static int last_vm_running;
static int gui_saved_grab;
static int gui_fullscreen;
static int gui_key_modifier_pressed;
static int gui_keysym;
static int gui_fullscreen_initial_grab;
static int gui_grab_code = gtkaltleft | gtkcontrolleft;
static uint8_t modifiers_state[256];
static int window_resize = 0;
static unsigned int cw, ch;
static gint win_x, win_y, win_w, win_h;
static GtkWidget * box; /* because GtkWindow/GtkBin holds only one widget */

static gboolean gtk2_expose(GtkWidget *wid, GdkEventExpose *event)
{
    if (gui_fullscreen && wid == event_port) return FALSE;
    if (!gui_fullscreen && wid == window) return FALSE;
    gdk_draw_image(wid->window, wid->style->fg_gc[GTK_WIDGET_STATE(wid)], image, event->area.x, event->area.y, event->area.x, event->area.y, event->area.width, event->area.height);
    return TRUE;
}

static void gtk2_update(DisplayState *ds, int x, int y, int w, int h)
{
    //    printf("updating x=%d y=%d w=%d h=%d\n", x, y, w, h);
    GdkEventExpose ev;
    ev.area.x = x;
    ev.area.y = y;
    ev.area.width = w;
    ev.area.height = h;
    /* catch the first resize done by the init function - make sure we don't **
    ** try to draw an image until screen has been finalized/realized/etc */
    if (event_port->window != NULL)
    {
	if (gui_fullscreen)
	    gtk2_expose(window, &ev);
	else
	    gtk2_expose(event_port, &ev);
    }
}

static void gtk2_resize(DisplayState *ds, int w, int h)
{

    //    printf(" resizing to %d %d\n", w, h);

    if (gui_fullscreen)
    {
	if (cw != w || ch != h)
	{
	    fullscreen_switch(win_x, win_y, w, h); /* changing video modes */
/* apparently GTK enjoys moving static windows all over the place while
they are being resized - so we have to be tricky */
	    //gtk_window_move(GTK_WINDOW(window), win_x, win_y);
	    //gtk_window_move(GTK_WINDOW(window), 0, 0);
	    gdk_window_move_resize(window->window, 0, 0, w, h);
	}
    }

    cw = w; ch = h;

    if (image)
	 g_object_unref(image);

/* gdk_visual_get_best_with_depth() ??? but then how to paint onto window? */
    image = gdk_image_new(GDK_IMAGE_NORMAL, gdk_visual_get_system(), w, h);
    gdk_image_set_colormap(image, gdk_colormap_get_system());

    gtk_widget_set_size_request(event_port, w, h);
    if (window_resize)
	gdk_window_move_resize(window->window, win_x, win_y, w+25, h+50);

    ds->data = image->mem;
    ds->linesize = image->bpl;
    ds->depth = image->bits_per_pixel;
    ds->width = w;
    ds->height = h;
    gtk2_update(ds, 0, 0, w, h);
}

static gboolean gtk2_config(GtkWidget *wid, GdkEventConfigure *ev, DisplayState *ds)
{
    if (!gui_fullscreen)
    {
	win_x = ev->x;
	win_y = ev->y;
	win_w = ev->width;
	win_h = ev->height;
    }
    return FALSE;
}

/* generic keyboard conversion */

#include "gdk_keysym.h"
#include "keymaps.c"

static kbd_layout_t *kbd_layout = NULL;

static uint8_t gtk2_keyevent_to_keycode_generic(const GdkEventKey *ev)
{
    int keysym;
    /* workaround for X11+SDL bug with AltGR - is it still needed for Gtk2? */
    keysym = ev->keyval;
    if (keysym == 0 && ev->hardware_keycode == 113)
	keysym = GDK_Mode_switch;
    return keysym2scancode(kbd_layout, keysym);
}

/* specific keyboard conversions from scan codes */

#if defined(_WIN32)

#include <windows.h>

static UINT vk2scan(UINT vk)
{
	return MapVirtualKey(vk,0);
}
static uint8_t gtk2_keyevent_to_keycode(const GdkEventKey *ev)
{   
    return (uint8_t)vk2scan((UINT)(ev->hardware_keycode));
}

#else

static const uint8_t x_keycode_to_pc_keycode[61] = {
   0xc7,      /*  97  Home   */
   0xc8,      /*  98  Up     */
   0xc9,      /*  99  PgUp   */
   0xcb,      /* 100  Left   */
   0x4c,        /* 101  KP-5   */
   0xcd,      /* 102  Right  */
   0xcf,      /* 103  End    */
   0xd0,      /* 104  Down   */
   0xd1,      /* 105  PgDn   */
   0xd2,      /* 106  Ins    */
   0xd3,      /* 107  Del    */
   0x9c,      /* 108  Enter  */
   0x9d,      /* 109  Ctrl-R */
   0x0,       /* 110  Pause  */
   0xb7,      /* 111  Print  */
   0xb5,      /* 112  Divide */
   0xb8,      /* 113  Alt-R  */
   0xc6,      /* 114  Break  */   
   0xdb,         /* 115 windows left button */
   0xdc,         /* 116 windows right button */
   0xdd,         /* 117 right menu button */
   0x0,         /* 118 */
   0x0,         /* 119 */
   0x70,         /* 120 Hiragana_Katakana */
   0x0,         /* 121 */
   0x0,         /* 122 */
   0x73,         /* 123 backslash */
   0x0,         /* 124 */
   0x0,         /* 125 */
   0x0,         /* 126 */
   0x0,         /* 127 */
   0x0,         /* 128 */
   0x79,         /* 129 Henkan */
   0x0,         /* 130 */
   0x7b,         /* 131 Muhenkan */
   0x0,         /* 132 */
   0x7d,         /* 133 Yen */
   0x0,         /* 134 */
   0x0,         /* 135 */
   0x47,         /* 136 KP_7 */
   0x48,         /* 137 KP_8 */
   0x49,         /* 138 KP_9 */
   0x4b,         /* 139 KP_4 */
   0x4c,         /* 140 KP_5 */
   0x4d,         /* 141 KP_6 */
   0x4f,         /* 142 KP_1 */
   0x50,         /* 143 KP_2 */
   0x51,         /* 144 KP_3 */
   0x52,         /* 145 KP_0 */
   0x53,         /* 146 KP_. */
   0x47,         /* 147 KP_HOME */
   0x48,         /* 148 KP_UP */
   0x49,         /* 149 KP_PgUp */
   0x4b,         /* 150 KP_Left */
   0x4c,         /* 151 KP_ */
   0x4d,         /* 152 KP_Right */
   0x4f,         /* 153 KP_End */
   0x50,         /* 154 KP_Down */
   0x51,         /* 155 KP_PgDn */
   0x52,         /* 156 KP_Ins */
   0x53,         /* 157 KP_Del */
};

static uint8_t gtk2_keyevent_to_keycode(const GdkEventKey *ev)
{
    int keycode;

    keycode = ev->hardware_keycode;

    if (keycode < 9) {
	keycode = 0;
    } else if (keycode < 97) {
	keycode -= 8; /* just an offset */
    } else if (keycode < 158) {
	/* use conversion table */
	keycode = x_keycode_to_pc_keycode[keycode - 97];
    } else {
	keycode = 0;
    }
    return keycode;
}

#endif

static void reset_keys(void)
{
    int i;
    for(i = 0; i < 256; i++) {
	if (modifiers_state[i]) {
	    if (i & 0x80)
		kbd_put_keycode(0xe0);
	    kbd_put_keycode(i | 0x80);
	    modifiers_state[i] = 0;
	}
    }
}

/* convert GDK modifiers and invoke ugly hack to distinguish
between left and right shift/control/alt */
static guint gtk2_GetModState(const GdkEventKey *ev)
{
	guint key = 0, keyval = ev->keyval, state = ev->state;
	switch(keyval)
	{
		case GDK_Shift_L:
			if (ev->type != GDK_KEY_RELEASE)
				key |= gtkshiftleft;
			keyval = 1;
			break;
		case GDK_Shift_R:
			if (ev->type != GDK_KEY_RELEASE)
				key |= gtkshiftright;
			keyval = 1;
			break;
		case GDK_Control_L:
			if (ev->type != GDK_KEY_RELEASE)
				key |= gtkcontrolleft;
			keyval = 2;
			break;
		case GDK_Control_R:
			if (ev->type != GDK_KEY_RELEASE)
				key |= gtkcontrolright;
			keyval = 2;
			break;
		case GDK_Alt_L:
			if (ev->type != GDK_KEY_RELEASE)
				key |= gtkaltleft;
			keyval = 3;
			break;
		case GDK_Alt_R:
			if (ev->type != GDK_KEY_RELEASE)
				key |= gtkaltright;
			keyval = 3;
			break;
		case GDK_Caps_Lock:
			if (ev->type != GDK_KEY_RELEASE)
				key |= gtkcapslock;
			keyval = 4;
			break;
		default:
			keyval = 0;
			break;
	}
	if (keyval != 1 && (state & GDK_SHIFT_MASK))
	{
		if (modifiers_state[0x2a])
			key |= gtkshiftleft;
		if (modifiers_state[0x36])
			key |= gtkshiftright;
	}
	if (keyval != 2 && (state & GDK_CONTROL_MASK))
	{
		if (modifiers_state[0x1d])
			key |= gtkcontrolleft;
		if (modifiers_state[0x9d])
			key |= gtkcontrolright;
	}
	if (keyval != 3 && (state & GDK_MOD1_MASK)) /* fixme: need to do a check to make sure that alt is mapped to GDK_MOD1_MASK in the GDK_Keymap */
	{
		if (modifiers_state[0x38])
			key |= gtkaltleft;
		if (modifiers_state[0xb8])
			key |= gtkaltright;
	}
	if (keyval != 4 && (state & GDK_LOCK_MASK))
		key |= gtkcapslock;
	return key;
}

static void gtk2_process_key(GdkEventKey *ev)
{
    int keycode, v;

    if (ev->keyval == GDK_Pause) {
	/* specific case */
	v = 0;
	if (ev->type == GDK_KEY_RELEASE)
	    v |= 0x80;
	kbd_put_keycode(0xe1);
	kbd_put_keycode(0x1d | v);
	kbd_put_keycode(0x45 | v);
	return;
    }

    if (kbd_layout) {
	keycode = gtk2_keyevent_to_keycode_generic(ev);
    } else {
	keycode = gtk2_keyevent_to_keycode(ev);
    }

    switch(keycode) {
    case 0x00:
	/* sent when leaving window: reset the modifiers state */
	reset_keys();
	return;
    case 0x2a:                          /* Left Shift */
    case 0x36:                          /* Right Shift */
    case 0x1d:                          /* Left CTRL */
    case 0x9d:                          /* Right CTRL */
    case 0x38:                          /* Left ALT */
    case 0xb8:                         /* Right ALT */
	if (ev->type == GDK_KEY_RELEASE)
	    modifiers_state[keycode] = 0;
	else
	    modifiers_state[keycode] = 1;
	break;
    case 0x45: /* num lock */
    case 0x3a: /* caps lock */
	/* GTK does send the key up event, so we dont generate it */
	/*kbd_put_keycode(keycode);
	kbd_put_keycode(keycode | 0x80);
	return;*/
	break;
    }

    /* now send the key code */
    if (keycode & 0x80)
	kbd_put_keycode(0xe0);
    if (ev->type == GDK_KEY_RELEASE)
	kbd_put_keycode(keycode | 0x80);
    else
	kbd_put_keycode(keycode & 0x7f);
}

static void gtk2_update_caption(void)
{
    char buf[1024];
    strcpy(buf, "QEMU Gtk");
    if (!vm_running) {
	strcat(buf, " [Stopped]");
    }
    if (gui_grab) {
	strcat(buf, " - Press Ctrl-Alt to exit grab");
    }
    gtk_window_set_title(GTK_WINDOW(window), buf);
}

/* what a nasty hack. this should be a part of the GDK, not external!!! */
#include "gdk_set_window_pointer.c"

static void gtk2_grab_start(void)
{
    gint y;
    guint events;
    GdkModifierType state; /* dummy var */
    GtkWidget * grab;
    if (gui_fullscreen) /* make sure grab is on correct x window */
    	grab = window;
    else
	grab = event_port;
if (gui_rgrab || gui_fullscreen)
{
    events = GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK | GDK_SCROLL_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK;
    y = gdk_pointer_grab(grab->window, TRUE, events, grab->window, invisible_cursor, GDK_CURRENT_TIME);
    if (y)
	printf("GTK Warning - pointer grab failed!\n");
    y = gdk_keyboard_grab(grab->window, TRUE, GDK_CURRENT_TIME);
    if (y)
	printf("GTK Warning - keyboard grab failed!\n");
}
    /* do a dummy read to avoid moving mouse - set ox and oy to stay in sync */
    gdk_window_get_pointer(event_port->window, &ox, &oy, &state);
    gui_grab = 1;
    gtk2_update_caption();
}

static void gtk2_grab_end(void)
{
    GtkWidget * grab;
    if (gui_fullscreen)
    	grab = window;
    else
	grab = event_port;
if (gui_rgrab || gui_fullscreen)
{
    gdk_pointer_ungrab(GDK_CURRENT_TIME);
    gdk_keyboard_ungrab(GDK_CURRENT_TIME);
    gdk_window_set_pointer(grab->window, win_x, win_y);
}
    gui_grab = 0;
    gtk2_update_caption();
}

static gboolean gtk2_send_mouse_event(GtkWidget *wid, GdkEvent *ev)
{
    int x, y, dx, dy, dz, state, buttons;
    GtkWidget * grab;
    if (gui_fullscreen) /* make sure grab is on correct x window */
    	grab = window;
    else
	grab = event_port;

    if (gui_fullscreen && wid == event_port) return FALSE;
    if (!gui_fullscreen && wid == window) return FALSE;
    if (ev->type == GDK_2BUTTON_PRESS || ev->type == GDK_3BUTTON_PRESS)
	return TRUE; /* double or triple click - superflurious - ignore */

    if (gui_grab)
    {

    if (ev->type == GDK_MOTION_NOTIFY && ev->motion.is_hint)
    {
	gdk_window_get_pointer(ev->motion.window, &x, &y, (GdkModifierType*)&state);
    }
    else
    {
	x = ev->motion.x;
	y = ev->motion.y;
	/* scroll.state occupies a different position in the union */
	if (ev->type == GDK_SCROLL)
	state = ev->scroll.state;
	else
	state = ev->motion.state;
    }

    dx = x - ox;
    dy = y - oy;
    // prevent infinite looping - 2.6.X
    if ((ev->type == GDK_MOTION_NOTIFY) && (dx == 0) && (dy == 0)) return TRUE;
    dz = 0;
    ox = x;
    oy = y;

    buttons = 0;
    if ((state & GDK_BUTTON1_MASK))
	buttons |= MOUSE_EVENT_LBUTTON;
    if ((state & GDK_BUTTON3_MASK))
	buttons |= MOUSE_EVENT_RBUTTON;
    if ((state & GDK_BUTTON2_MASK))
	buttons |= MOUSE_EVENT_MBUTTON;

    if (ev->type == GDK_BUTTON_PRESS)
    {
	if (ev->button.button == 1)
	    buttons |= MOUSE_EVENT_LBUTTON;
	if (ev->button.button == 3)
	    buttons |= MOUSE_EVENT_RBUTTON;
	if (ev->button.button == 2)
	    buttons |= MOUSE_EVENT_MBUTTON;
    }
    else if (ev->type == GDK_BUTTON_RELEASE)
    {
    /* not sure if this is really necessary, but just to be on the safe side **
    ** reset qemu's mask so that a button thats being released will be shown **
    * missing from the mask (which lets the guest know the button was relased */
	buttons = 0;

	if ((state & GDK_BUTTON1_MASK) && ev->button.button != 1)
	    buttons |= MOUSE_EVENT_LBUTTON;
	if ((state & GDK_BUTTON3_MASK) && ev->button.button != 3)
	    buttons |= MOUSE_EVENT_RBUTTON;
	if ((state & GDK_BUTTON2_MASK) && ev->button.button != 2)
	    buttons |= MOUSE_EVENT_MBUTTON;
    }
    else if (ev->type == GDK_SCROLL)
    {
	/* test wheel - copied from Sebastien Bechet's gtk.c */
	if (ev->scroll.direction == GDK_SCROLL_UP)
	    dz--;
	if (ev->scroll.direction == GDK_SCROLL_DOWN)
	    dz++;
    }

    if ((ev->type == GDK_MOTION_NOTIFY) && (gui_rgrab || gui_fullscreen))
    {
	/* wrap the x,y coordinates back onto the window */
	if (ev->motion.x <= (cw/4))
	    x = ((3*cw/4)-1);
	if (ev->motion.y <= (ch/4))
	    y = ((3*ch/4)-1);
	if (ev->motion.x >= (3*cw/4))
	    x = (cw/4)+1;
	if (ev->motion.y >= (3*ch/4))
	    y = (ch/4)+1;

	/* make internal mouse move invisible */
	ox = x;
	oy = y;

	gdk_window_set_pointer(grab->window, (gint)x, (gint)y);
    }

    kbd_mouse_event(dx, dy, dz, buttons);

    }
    else
    {

    if (ev->type == GDK_BUTTON_PRESS && ev->button.button == 1)
    {
	/* start grabbing all events */
	gtk2_grab_start();
    }

    }
    return TRUE;
}

static void toggle_full_screen(DisplayState *ds)
{
    gui_fullscreen = !gui_fullscreen;
    gtk2_resize(ds, image->width, image->height);
    if (gui_fullscreen) {
	gui_saved_grab = gui_grab;
	gtk2_grab_start();
	gtk_window_get_position(GTK_WINDOW(window), &win_x, &win_y);
	gdk_window_move_resize(window->window, 0, 0, ds->width, ds->height);
	gtk_widget_hide(box);
	fullscreen_switch(win_x, win_y, ds->width, ds->height);
    } else {
	fullscreen_reset();
	gdk_window_move_resize(window->window, win_x, win_y, win_w, win_h);
	gui_fullscreen = !gui_fullscreen;
	if (!gui_saved_grab)
	    gtk2_grab_end();
	gui_fullscreen = !gui_fullscreen;
	gtk_widget_show(box);
    }
    vga_invalidate_display();
    vga_update_display();
}

static gboolean gtk2_key_press(GtkWidget *wid, GdkEventKey *ev, DisplayState *ds)
{
	int mod_state, internal_key = 0;
	    if (ev->type == GDK_KEY_PRESS) {
		mod_state = (gtk2_GetModState(ev) & (int)gui_grab_code) == (int)gui_grab_code;
		gui_key_modifier_pressed = mod_state;
		if (gui_key_modifier_pressed) {
		    int keycode;
		    internal_key = 1;
		    keycode = gtk2_keyevent_to_keycode(ev);
		    switch(keycode) {
		    case 0x21: /* 'f' key on US keyboard */
			toggle_full_screen(ds);
			gui_keysym = 1;
			break;
		    case 0x02 ... 0x0a: /* '1' to '9' keys */ 
			console_select(keycode - 0x02);
			if (is_active_console(vga_console)) {
			    /* tell the vga console to redisplay itself */
			    vga_invalidate_display();
			} else {
			    /* display grab if going to a text console */
			    if (gui_grab)
				gtk2_grab_end();
			}
			gui_keysym = 1;
			break;
		    default:
			break;
		    }
		} else if (!is_active_console(vga_console)) {
		    int keysym;
		    keysym = 0;
		    if (ev->state & GDK_CONTROL_MASK) {
			switch(ev->keyval) {
			case GDK_Up: keysym = QEMU_KEY_CTRL_UP; break;
			case GDK_Down: keysym = QEMU_KEY_CTRL_DOWN; break;
			case GDK_Left: keysym = QEMU_KEY_CTRL_LEFT; break;
			case GDK_Right: keysym = QEMU_KEY_CTRL_RIGHT; break;
			case GDK_Home: keysym = QEMU_KEY_CTRL_HOME; break;
			case GDK_End: keysym = QEMU_KEY_CTRL_END; break;
			case GDK_Page_Up: keysym = QEMU_KEY_CTRL_PAGEUP; break;
			case GDK_Page_Down: keysym = QEMU_KEY_CTRL_PAGEDOWN; break;
			default: break;
			}
		    } else {
			switch(ev->keyval) {
			case GDK_Up: keysym = QEMU_KEY_UP; break;
			case GDK_Down: keysym = QEMU_KEY_DOWN; break;
			case GDK_Left: keysym = QEMU_KEY_LEFT; break;
			case GDK_Right: keysym = QEMU_KEY_RIGHT; break;
			case GDK_Home: keysym = QEMU_KEY_HOME; break;
			case GDK_End: keysym = QEMU_KEY_END; break;
			case GDK_Page_Up: keysym = QEMU_KEY_PAGEUP; break;
			case GDK_Page_Down: keysym = QEMU_KEY_PAGEDOWN; break;
			case GDK_BackSpace: keysym = QEMU_KEY_BACKSPACE; break;
			case GDK_Delete: keysym = QEMU_KEY_DELETE; break;
			default: break;
			}
		    }
		    if (keysym) {
			kbd_put_keysym(keysym);
		    } /*else if (ev->key.keysym.unicode != 0) {
			kbd_put_keysym(ev->key.keysym.unicode);
		    }*/
		}
	    } else if (ev->type == GDK_KEY_RELEASE) {
		mod_state = (gtk2_GetModState(ev) & gui_grab_code);
		if (!mod_state) {
		    if (gui_key_modifier_pressed) {
			if (gui_keysym == 0) {
			    /* exit/enter grab if pressing Ctrl-Alt */
			    if (!gui_grab)
				gtk2_grab_start();
			    else
				gtk2_grab_end();
			    /* SDL does not send back all the
			       modifiers key, so we must correct it */
			    reset_keys();
			    return TRUE;
			}
			gui_key_modifier_pressed = 0;
			gui_keysym = 0;
			internal_key = 1;
		    }
		}
	    }
	    if (is_active_console(vga_console) && !internal_key) 
		gtk2_process_key(ev);
            else if (internal_key && ev->type == GDK_KEY_RELEASE)
            {
                modifiers_state[0x38] = 0;
                modifiers_state[0x1d] = 0;
                kbd_put_keycode(0x38 | 0x80);
                kbd_put_keycode(0x1d | 0x80);
            }
            else if (internal_key && ev->type == GDK_KEY_PRESS)
            {
                modifiers_state[0x38] = 1;
                modifiers_state[0x1d] = 1;
            }
	return TRUE;
}

static void gtk2_refresh(DisplayState *ds)
{
    if (last_vm_running != vm_running) {
	last_vm_running = vm_running;
	gtk2_update_caption();
    }
    if (ds->data != image->mem)
    {
	ds->data = image->mem;
    }

    if (is_active_console(vga_console))                                         
	vga_update_display();                                                   
    while (gtk_events_pending())
	gtk_main_iteration();
}

static void gtk2_cleanup(void) 
{
    if (gtk_main_level() != 0)
	gtk_main_quit();
    fullscreen_cleanup();
}

static gboolean gtk2_deletewin(void)
{
    /* signal qemu that its time to shut itself off - this is the place that we hook to trap attempts to close the main qemu window */
    qemu_system_shutdown_request();
    return FALSE;
    return TRUE; /* dont close the window right away! give qemu time to think */
}

static void gtk2_destroy(void)
{
    /* ideally we would call a hook here so qemu could clean itself up */
    gtk2_cleanup();
}

GtkWidget * gui_menu_add(const gchar * name)
{
    GtkWidget *menu, *menu_name;
    menu = gtk_menu_new();
    menu_name = gtk_menu_item_new_with_label(name);
    gtk_menu_shell_append(GTK_MENU_SHELL(menubar), menu_name);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_name), menu);
    gtk_widget_show(menu_name);
    return menu;
}

GtkWidget * gui_menuitem_add(GtkWidget * menu, const gchar * name, void (*item_handler)(void))
{
    GtkWidget *menu_item;
    menu_item = gtk_menu_item_new_with_label(name);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
    g_signal_connect(G_OBJECT(menu_item), "activate", G_CALLBACK(item_handler), NULL);
    gtk_widget_show(menu_item);
    return menu_item;
}

GtkWidget * gui_menusep_add(GtkWidget * menu)
{
    GtkWidget *menu_item;
    menu_item = gtk_menu_item_new();
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item);
    gtk_widget_show(menu_item);
    return menu_item;
}

GtkWidget * gui_submenu_add(GtkWidget * menu, const gchar * name)
{
    GtkWidget *submenu, *menu_name;
    submenu = gtk_menu_new();
    menu_name = gtk_menu_item_new_with_label(name);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_name);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_name), submenu);
    gtk_widget_show(menu_name);
    return submenu;
}

void base_vc_switch(int i)
{
	console_select(i);
	if (is_active_console(vga_console)) {
	    /* tell the vga console to redisplay itself */
	    vga_invalidate_display();
	} else {
	    /* display grab if going to a text console */
	    if (gui_grab)
		gtk2_grab_end();
	}
}
void vc_guest(void) { base_vc_switch(0); }
void vc_monitor(void) { base_vc_switch(1); }
void vc_serial(void) { base_vc_switch(2); }
void vc_para(void) { base_vc_switch(3); }
void nop(void) {}

void gtk2_menu_init(void)
{
	GtkWidget * menu;
	menu = gui_menu_add("Monitor");
	gui_menuitem_add(menu, "Reset", qemu_system_reset_request);
	gui_menuitem_add(menu, "Quit", qemu_system_shutdown_request);
	menu = gui_menu_add("Virtual Console");
	menu = gui_submenu_add(menu, "Switch to");
	gui_menuitem_add(menu, "Guest", vc_guest);
	gui_menuitem_add(menu, "Monitor", vc_monitor);
	gui_menuitem_add(menu, "Serial", vc_serial);
	gui_menuitem_add(menu, "Parallel", vc_para);
	menu = gui_menu_add("Help");
	gui_menuitem_add(menu, "About", nop);
}

static int gargc=0;
static char ** gargv=NULL;
void gtk2_display_init(DisplayState *ds, int full_screen)
{
    int events;

#if defined(__APPLE__)
    /* always use generic keymaps */
    if (!keyboard_layout)
	keyboard_layout = "en-us";
#endif
    if(keyboard_layout) {
	kbd_layout = init_keyboard_layout(keyboard_layout);
	if (!kbd_layout)
	    exit(1);
    }

/* note - this strips the GTK-specific args after the main program is done with them */
    if (!gtk_init_check (&gargc,&gargv))
    {
	fprintf(stderr, "Could not load GTK\n");
	exit(0);
    }
    gui_rgrab = 0;
    fullscreen_init();

/* note: adding GDK_DRAG_* and GDK_DROP_* would provide a mechanism for supporting drag and drop between host and guest */
    events = GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_FOCUS_CHANGE_MASK | GDK_SCROLL_MASK | GDK_STRUCTURE_MASK;

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    swindow = gtk_scrolled_window_new(NULL,NULL);
    event_port = gtk_event_box_new();
    menubar = gtk_menu_bar_new();
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swindow), GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);
    box = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(window), box);
    gtk_box_pack_start(GTK_BOX(box), menubar, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(box), swindow, TRUE, TRUE, 0);
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(swindow), event_port);
    gtk_widget_set_events(event_port, events);
    gtk_window_set_default_size(GTK_WINDOW(window), 720, 420);
    gtk_widget_set_size_request(window, 0, 0);
    gtk_widget_set_size_request(swindow, 720, 400);
    gtk_widget_set_size_request(event_port, 720, 400);
    gtk_window_set_gravity(GTK_WINDOW(window), GDK_GRAVITY_STATIC);
    gtk2_menu_init();

    g_signal_connect(G_OBJECT(window), "delete_event", G_CALLBACK(gtk2_deletewin), NULL);
    g_signal_connect(G_OBJECT(window), "destroy", G_CALLBACK(gtk2_destroy), NULL);
    gtk_container_set_border_width(GTK_CONTAINER(window), 0);

    g_signal_connect(G_OBJECT(window), "configure_event", G_CALLBACK(gtk2_config), NULL);
    g_signal_connect(G_OBJECT(window), "expose_event", G_CALLBACK(gtk2_expose), NULL);
    g_signal_connect(G_OBJECT(window), "motion_notify_event", G_CALLBACK(gtk2_send_mouse_event), NULL);
    g_signal_connect(G_OBJECT(window), "button_press_event", G_CALLBACK(gtk2_send_mouse_event), NULL);
    g_signal_connect(G_OBJECT(window), "button_release_event", G_CALLBACK(gtk2_send_mouse_event), NULL);
    g_signal_connect(G_OBJECT(window), "scroll_event", G_CALLBACK(gtk2_send_mouse_event), NULL);
    g_signal_connect(G_OBJECT(event_port), "expose_event", G_CALLBACK(gtk2_expose), NULL);
    g_signal_connect(G_OBJECT(event_port), "motion_notify_event", G_CALLBACK(gtk2_send_mouse_event), NULL);
    g_signal_connect(G_OBJECT(event_port), "button_press_event", G_CALLBACK(gtk2_send_mouse_event), NULL);
    g_signal_connect(G_OBJECT(event_port), "button_release_event", G_CALLBACK(gtk2_send_mouse_event), NULL);
    g_signal_connect(G_OBJECT(event_port), "scroll_event", G_CALLBACK(gtk2_send_mouse_event), NULL);
    g_signal_connect(G_OBJECT(window), "key_press_event", G_CALLBACK(gtk2_key_press), ds);
    g_signal_connect(G_OBJECT(window), "key_release_event", G_CALLBACK(gtk2_key_press), ds);

    ds->dpy_update = gtk2_update;
    ds->dpy_resize = gtk2_resize;
    ds->dpy_refresh = gtk2_refresh;

    gchar nullpixdata[1] = { 0 };
    GdkColor nullcolor = { 0, 0, 0, 0 };
    GdkPixmap *invis = gdk_bitmap_create_from_data(window->window, nullpixdata, 1, 1);
    invisible_cursor = gdk_cursor_new_from_pixmap(invis, invis, &nullcolor, &nullcolor, 0, 0);

    gtk2_resize(ds, 720, 400);
    gtk2_update_caption();
    gui_grab = 0;

    gtk_widget_show(event_port);
    gtk_widget_show(swindow);
    gtk_widget_show(menubar);
    gtk_widget_show(box);
    gtk_widget_show(window);
    if (full_screen) {
	gui_fullscreen = 1;
	gui_fullscreen_initial_grab = 1;
	gtk2_grab_start();
    }
}

/* we are allowed to look at the args, but not allowed to modify them
with the exception that we have to remove args that are meant for us, as
main qemu will not recognize them */
void gui_checkargs(int * pargc, char *** pargv)
{
    int argc = *pargc;
    char ** argv = *pargv;
    char ** vargv;
    int i, j = 0;
    vargv = malloc(argc * sizeof(char*));
    for (i = 0; i < argc; i++)
    {
        if (!strcmp(argv[i], "-window-resize-on"))
	{
		window_resize = 1;
		(*pargc)--;
	}
        else if (!strcmp(argv[i], "-raw-grab"))
	{
		gui_rgrab = 1;
		(*pargc)--;
	}
	else
	{
		vargv[j] = argv[i];
		j++;
	}
    }
    /* so lets hide them */
    for (i = 0; i < *pargc; i++)
    {
	(*pargv)[i] = vargv[i];
    }
    if (argc != (*pargc)) (*pargv)[*pargc] = NULL;
    free(vargv);
    /* save a copy so we can pass them to gtk_init() later */
    gargc = *pargc;
    gargv = *pargv;
}

